/*
 * Environment.java - DynamicBees, a demo for the probe and watcher mechanisms
 * Copyright (C) 1998-2002 Olivier Gutknecht, Fabien Michel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
//package dynamicbees;

import madkit.kernel.ReferenceableAgent;
import madkit.kernel.Watcher;
import java.awt.Color;

/**
  @version 2.0
  @author Fabien MICHEL 01/02/2001*/
public class Environment extends Watcher implements ReferenceableAgent
{
	// requis pour la sérialisation (!?!), 3* -> envirt
	private static final long serialVersionUID = 32L;

	public static final double timeStep = .1, speedUp = 100;
	public static final Color  ghostColor = new Color(.25f, .25f, .25f);
	public double xMax = 300, yMax = 300;
	RobotEnumerator robotEnumerator;
	// influence handling
	int inflReceived, stepNb; 
	double score; 
	boolean[] swarm; 
	Vector2D[] influences;

	public Environment(final double xSz, final double ySz, final int stepNb)
	{
		xMax = xSz;
		yMax = ySz; 
		this.stepNb = stepNb; 
		Robot.robotWorld = this;
	}

	public boolean testing() { return (stepNb > 0); }
	
	public double testResult() { return score; }

	public void updateRobots()
	{
		Robot[] rob = robotEnumerator.getRobots(), lead = new Robot[rob.length], 
				alone = new Robot[rob.length];
		int i, nl = 0, na = 0;
		// looking for robots not yet followers and for leaders
		for (i = 0; i < rob.length; i++)
			if (rob[i].color() == null)  { alone[na] = rob[i]; na++; }
			else if ( rob[i].leader() )  { lead[nl]  = rob[i]; nl++; }
		// make free robots follow a leader
		if (nl > 0)
			for (i = 0; i < na; i++)
			{ 
				Robot leader = lead[( (int) (Math.random()*nl) )];
				// the alone robot takes the color of its leader
				alone[i].setColor( leader.color() ); 
				// the alone robot is placed near the leader (or one of its followers)
				alone[i].setConfig( placeNear(leader) );
			}
		// update the influences
		updateInfluences(rob.length);
	}
	
	private OrientPoint placeNear(Robot robot) {
		final boolean debug = false;  // activates the console output
		final java.io.PrintStream out = System.out; 
		// find the robot's neighbours
		final java.util.Vector<Integer> ngbrs = getNeighbours(robot, true);
		final double dVar = Robot.Rpercept - Robot.diameter, 
				dMin = Robot.diameter + 2 * dVar / 3, dist = dMin + Math.random() * dVar / 6;
		OrientPoint res = null;
		if (debug) { 
			out.print("=== Robots # near "); robot.config().print(out); out.print(" : "); 
			for(int id = 0; id < ngbrs.size(); id++) 
			{ out.print( ngbrs.get(id) ); out.print(" "); }
			System.out.print("\n");
		}
		if (ngbrs.size() == 0) {
		    final double thVar = (Math.PI - Robot.Apercept - Robot.radius / dist) / 2, 
					theta = (2 * Math.random() - 1) * thVar;
			res = new OrientPoint(Vector2D.fromPolar(-dist, theta),
					      (2 * Math.random() - 1) * theta / 2); 
			res = robot.config().inFrame(res);
			if (debug) 
			{ out.print("    New robot "); res.print(out); out.print(" ===\n"); } 
		} else 
			res = placeNear( robotEnumerator.getRobots()[ngbrs.get(0)] );
		return res;
	}

	public void updateInfluences(final int nb)
	{
		Vector2D[] newInfluences = new Vector2D[nb];
		for(int iRob = 0; iRob < nb; iRob++)
			newInfluences[iRob] = (influences != null) && (iRob < influences.length)
			? influences[iRob] : null;
		influences = newInfluences;
		
		boolean[] newSwarm = new boolean[nb];
		for(int iRob = 0; iRob < nb; iRob++)
			newSwarm[iRob] = (swarm != null) && (iRob < swarm.length) ? swarm[iRob] : false;
		swarm = newSwarm;
	}

	public void activate()
	{
		createGroup(false,"robots",null,null);
		requestRole("react","robot environment",null);
		robotEnumerator=new RobotEnumerator("robots", "robot", this);
		addProbe(robotEnumerator);
		inflReceived = 0;
		influences   = null;
		updateRobots();
	}

	public void end()
	{
		disposeMyGUI();
		super.end();
	}

	public java.util.Vector<Integer> getNeighbours(final Robot robot, 
												   final boolean back) {
		final Robot[]  robots = robotEnumerator.getRobots();
		final OrientPoint cfg = robot.config();
		final double   orient = cfg.orientation(), 
				percRad = Robot.Rpercept / Robot.radius + 2;
		final Vector2D pos = cfg.position(), dir = Vector2D.fromPolar(1, orient);
		final int      nbNbrs = (int)(percRad * percRad);
		int iRob; 
		java.util.Vector<Integer> iNbrs = new java.util.Vector<Integer>(nbNbrs);

		// we search for the neighbours indexes
		for(iRob = 0; iRob < robots.length; iRob++) {
			final Vector2D v = robots[iRob].config().position().minus(pos);
			final double dist = v.length(), proj = v.scalProd(dir);
			if ( (dist > 0) && (dist <= Robot.Rpercept) && 
					( ( back && (proj <= 0) ) 
					  || ( (!back) && (dist * Math.cos(Robot.Apercept) <= proj) ) ) )
				iNbrs.add(iRob);
		} 
		return iNbrs;
	}

	/** Gives the other robots perceived by a given one, as an array of detection
	 *  in the perceptive robot's frame.
	 * 
	 * @param id  The perceptive robot's identifier.
	 * @return    The array of the detections of the robot # id.
	 */
	public Detection[] getPerceptions(final int id) 
	{
		final boolean debug = false;  // activates the console output
		final java.io.PrintStream out = System.out;
		final Robot[]  robots = robotEnumerator.getRobots();
		final OrientPoint cfg = robots[id].config();
		final double   orient = cfg.orientation(), percRad = Robot.Rpercept / Robot.radius;
		final Vector2D    pos = cfg.position(), dir = Vector2D.fromPolar(1, orient);
		final double robX = pos.getX(), robY = pos.getY(), robTh = cfg.orientation(); 
		final double[][] ghostCoord = 
			// ghost = projection of the robot on the environment's boundary
			// { {robX, 0}, {robX, yMax}, {0, robY}, {xMax, robY} }; 
			// better: ghost = middle between robot & projection
			{ {robX, robY / 2}, {robX, (yMax + robY) / 2}, 
			  {robX / 2, robY}, {(xMax + robX) / 2, robY} }; 
		final double[] ghostOrient = {- robTh, Math.PI - robTh}; 
		// we built the neighbours' configurations array
		final java.util.Vector<Integer> iNbrs = getNeighbours(robots[id], false);
		swarm[id] = robots[id].move() && ( (iNbrs.size() > 0) || robots[id].leader() ); 
		// add one or two ghost neighbours to remain in the environment
		if (robX + Robot.Rpercept > xMax)  iNbrs.add(-1);
		else if (robX < Robot.Rpercept)    iNbrs.add(-2);
		if (robY + Robot.Rpercept > yMax)  iNbrs.add(-3);
		else if (robY < Robot.Rpercept)    iNbrs.add(-4);	
		final int nbNbrs = iNbrs.size();
		Detection[] perceptions = new Detection[nbNbrs];
		int iDetect = 0;
		if ( debug && (nbNbrs > 0) ) { 
			out.print("=== Configuration ("); out.print( pos.getX() ); 
			out.print(", "); out.print( pos.getY() ); 
			out.print(", "); out.print( OrientPoint.rad2deg(orient) ); 
			out.print(")\n    Perceptions :"); 
		} 
		java.util.Iterator<Integer> iter = iNbrs.iterator(); 
		while( iter.hasNext() ) {
			final int iRob = iter.next(); 
			final boolean perGhost = iRob < 0;
			final Robot rob = perGhost ? null : robots[iRob]; 
			final int   iGhost = perGhost ? 4 + iRob : -1; 
			final OrientPoint percPt  = perGhost 
					? new OrientPoint(ghostCoord[iGhost][0], ghostCoord[iGhost][1],
									  ghostOrient[iGhost/2]) : rob.config();
			final Color   percCol = perGhost ? ghostColor : rob.color();
			final boolean percLdr = perGhost ? false : rob.leader(); 
			//if ( (percPt != null) && (percCol != null) ) {
				perceptions[iDetect] = new Detection( cfg.toFrame(percPt), percCol,
													  percLdr ); 
				iDetect++; /*
				if (debug) {
					out.print(" "); percPt.position().print(out); out.print(" -> "); 
					perceptions[iDetect].config().position().print(out); 
				} 
			} else {
				final java.io.PrintStream err = System.err;
				err.print(" !!! INVALID PERCEPTION (# "); err.print(iRob);
				err.print(" / "); err.print(nbNbrs); err.print(", code "); 
				err.print(iRob); err.print(" / "); err.print(robots.length); 
				err.print(") !!!");
			} */
		}
		if ( debug && (nbNbrs > 0) )  out.println(" ===");
		return perceptions;
	}

	public void setInfluence(final int id, Vector2D acc)
	{
		if ( (0 <= id) && (id < influences.length) ) 
		{ influences[id] = acc; inflReceived++; }
		else {
			System.err.print("-=# !!! Abnormal id ("); System.err.print(id); 
			System.err.print(" / "); System.err.print(influences.length); 
			System.err.println(") !!! #=-"); 
		}

		final Robot[] robots = robotEnumerator.getRobots();
		int newScore = 0;
		boolean noCollision = true;
		if (inflReceived == robots.length) // all received, handle
		{
			// no real (collision) handling yet (to do)
			for(int iRob = 0; iRob < inflReceived; iRob++) { 
				if (influences[iRob] != null)
					robots[iRob].move(influences[iRob]);
				else {
					System.err.print("-=# !!! No influence for robot # "); 
					System.err.print(id); System.err.println(" !!! #=-"); 
				}
				if (noCollision) {
					if ( robots[iRob].enCollision() ) 
					{ noCollision = false;  newScore = -1; }
					else  newScore += swarm[iRob] ? 1 : 0; 
				}
			}
			score = newScore;
			//System.out.print("--- Score for step "); System.out.print(stepNb); 
			//System.out.print(" is "); System.out.println(score); 
			if (score < 0)  stepNb = 0;
			// reset influences array
			inflReceived = 0; // following lines are useless
			// influences   = null;
			// updateInfluences(robots.length);  
			if (stepNb > 0)  stepNb--;
			else if (stepNb < 0) { // AjS: pause in Scheduler does not seem to work
				try { Thread.sleep( (int)(timeStep * 1000 / speedUp) ); }
				catch (InterruptedException e) { System.err.print("Sleep interrupted!"); }
			}
		}
	}
}
