/** An oriented point in R^2, representing the configuration 
 *  of a robot.
 *
 * @author  A. Scheuer, UL, FST, DI and LORIA.
 */
public class OrientPoint {
    private Vector2D pos;
    private double   orient;

    /** Default constructor, defining a (0, 0, 0) oriented point. 
     * @see  #set(Vector2D, double) */
    public OrientPoint() { set(new Vector2D(), 0); }

    /** Standard constructor, defining a (x, y, th) oriented point.
     * @param x       the abscissa of the point,
     * @param y       the ordinate of the point,
     * @param orient  the orientation. 
     * @see  #set(Vector2D, double) */
    public OrientPoint(final double x, final double y, final double orient)
    { set(new Vector2D(x, y), orient); }

    /** Simple constructor, defining the oriented point by its position
     * and orientation.
     * @param pos     the point,
     * @param orient  the orientation. 
     * @see  #set(Vector2D, double) */
    public OrientPoint(final Vector2D pos, final double orient)
    { set(pos, orient); }

    /** Copy constructor.
     * @param op  an other oriented point. 
     * @see  #set(Vector2D, double) */
    public OrientPoint(final OrientPoint op)
    { set( op.position(), op.orientation() ); }

    
    /**  Brings an orientation back to the interval (-PI, PI] (-PI excluded).
     * @param orient  the original orientation.
     * @return        the corresponding orientation (modulo 2 PI) in (-PI, PI]. */
    public static double in2pi(final double orient) {
    	final double max = 2 * Math.PI;  double res = orient;
    	while (res > Math.PI)   res -= max;  while (res <= -Math.PI)  res += max;
    	return res;
    }
    
    /** Converts an orientation from radiant to degrees.
     * @param rad  the angle in radiant.
     * @return     the same angle in degrees. */
    static double rad2deg(final double rad) { return rad * 180 / Math.PI; }

    
    /** Descriptive method, giving the position of the oriented point.
     * @return  the position of the oriented point. */
    public Vector2D position() { return pos; }

    /** Descriptive method, giving the orientation of the oriented point.
     * @return  the orientation of the oriented point (in radiant). */
    public double orientation() { return orient; }

    /** Descriptive method, giving the orientation in degrees 
     * of the oriented point.
     * @return  the orientation of the oriented point (in degrees). */
    public double orientDeg() { return rad2deg(orient); }
    
    /** Descriptive method, projecting an oriented point into the frame 
     * defined by the current.
     *
     * This method is the inverse of @ref inFrame: for all <code>q1</code> 
     * and <code>q2</code>, you should have 
     * <code>q2 = q1.toFrame( q1.inFrame(q2) )</code>.
     *
     * @param   cfg  the oriented point projected into the current's frame.
     * @return  the projection of the given oriented point.
     */
    public OrientPoint toFrame(final OrientPoint cfg) {
    	final Vector2D pos = cfg.position().minus( position() );
    	return new OrientPoint( pos.rotate( -orientation() ),
			        cfg.orientation() - orientation() );
    }

    /** Descriptive method, computing the oriented point from its projection 
     * into the frame defined by the current.
     *
     * This method is the inverse of @ref toFrame: for all <code>q1</code> 
     * and <code>q2</code>, you should have 
     * <code>q2 = q1.inFrame( q1.toFrame(q2) )</code>.
     *
     * @param   cfg  the oriented point's projection into the current's frame.
     * @return  the oriented point whose projection is given.
     */
    public OrientPoint inFrame(final OrientPoint cfg) {
	final Vector2D turned = cfg.position().rotate( orientation() );
	OrientPoint res = new OrientPoint(this);
	res.add( turned, cfg.orientation() ); 
    	return res;
    }

    /** Descriptive method, printing the oriented point into a stream.
     * @param   out  the stream into which the oriented point will be printed. */
    public void print(java.io.PrintStream out) {
    	out.print("("); out.print( position().getX() ); 
    	out.print(", "); out.print( position().getY() ); 
    	out.print(", "); out.print( orientation() ); out.print(")");
    }
    
    /** Modification method, modifying the position and orientation 
     * of the oriented point.
     * @param pos     the position of the oriented point,
     * @param orient  the orientation of the oriented point. */
    public void set(final Vector2D pos, final double orient)
    { this.pos = new Vector2D(pos); this.orient = in2pi(orient); }

    /** Modification method, adding a position and an orientation 
     * to the oriented point.
     * @param pos     the vector added to the position of the oriented point,
     * @param orient  the angle added to the orientation of the oriented point. */
    public void add(final Vector2D pos, final double orient)
    { this.pos.add(pos); this.orient = in2pi(this.orient + orient); }
    
    
} // end of class OrientPoint 
