/*
* Robot.java - DynamicBees, a demo for the probe and watcher mechanisms
* Copyright (C) 1998-2002 Olivier Gutknecht, Fabien Michel
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
//package dynamicbees;

import java.awt.Color;

import madkit.kernel.AbstractAgent;
import madkit.kernel.ReferenceableAgent;
/**
  @version 2.0
  @author Olivier Gutknecht
  @author Fabien MICHEL 01/02/2001*/
public class Robot extends AbstractAgent implements ReferenceableAgent {
    // requis pour la sérialisation (!?!), 2* -> agents
    private static final long serialVersionUID = 21L;

    public enum eVectors {avoidance, cohesion, alignment, vectorsNb};
    private static double[] vectorsCoef;
    
    public static final double radius = 7, diameter = 2 * radius,
    // perception radius (>> 2 * diameter) and half-angle  
	Rpercept = 8 * radius, Apercept = 2 * Math.PI / 3, 
	// Avoidance radius (>> diameter)
	Ravoid = Rpercept / 2; 
    public static Environment robotWorld;
    
    public static double linVelMax = Robot.Rpercept / 3, rotVelMax = Math.PI / 12, 
    		accMax = Robot.linVelMax / 3, accMin = -1.5 * Robot.accMax;

    private int id, collision;
    private boolean isLeader, hasNeighbours;
    private OrientPoint config;
    private Vector2D vel;
    private Color robotColor;
  
    public Robot(final int id, boolean isLeader) {
		this.id = id;
		this.isLeader = isLeader;
		hasNeighbours = false;
		final double offset = diameter * 20, offset2 = 2 * offset; 
		config  = new OrientPoint(offset + Math.random() * (robotWorld.xMax - offset2),
				  				  offset + Math.random() * (robotWorld.yMax - offset2),
								  (Math.random() * 2 - 1) * Math.PI);
		vel     = new Vector2D();
		// avoiding pure red or green (always a minimum of 0.5 of blue)
	    robotColor = (isLeader ? new Color( (float) Math.random(), (float) Math.random(),
	    									(float)(Math.random() / 2 + .5) ) : null);
    }
    
    public Color color() { return robotColor; }
    public OrientPoint config() { return config; }
    public boolean leader() { return isLeader; }
    public boolean move() { return vel.getX() >= Robot.rotVelMax / 10; }
    public boolean neighbours() { return hasNeighbours; }
    public boolean enCollision() { return collision > 0; }

    public static void setCoefs(double[] coefs) 
    { if ( coefs.length == eVectors.vectorsNb.ordinal() )  vectorsCoef = coefs; }   
    
    public void setColor(Color col) { robotColor = col; } 
    public void setConfig(OrientPoint cfg) { config = cfg; } 
    public void setCollision() { collision = 5; }  
    
    public void activate() {
    	createGroup(true, "react", "robots", null, null);
    	requestRole("react", "robots", "robot", null);
    }
    
    public void react() { 
		final boolean debug = false;  // activates the console output
    	if (collision > 0)  collision--;  // timer for the collision color
		// compute the bounds of the accelerations
		final double linFact = vel.getX() < 0 ? -1 : 1, rotFact = vel.getY() < 0 ? -1 : 1, 
				linAccMax = Math.min(accMax, ( linVelMax - linFact * vel.getX() ) 
											 / Environment.timeStep), 
				rotMaxAcc = accMax / radius, rotMinAcc = accMin / radius,
				rotAccMax = Math.min(rotMaxAcc, ( rotVelMax - rotFact * vel.getY() )
							            		/ Environment.timeStep), 
				distLim = Ravoid - diameter, timeHoriz = 1.5 * Environment.timeStep, 
				maxFact = 25;
		// === compute the cohesion, avoidance & alignment vectors for perceptions ===
		Vector2D[] vectors = new Vector2D[eVectors.vectorsNb.ordinal()];
		for(int iVectors = 0; iVectors < eVectors.vectorsNb.ordinal(); iVectors++) 
			vectors[iVectors] = new Vector2D();
		// modify the vector according to the perceptions
    	Detection[] perceptions = robotWorld.getPerceptions(id);
    	hasNeighbours = (perceptions != null) && (perceptions.length > 0);
		if (debug) {
			System.err.print("  Lin. vel. = "); System.err.print( vel.getX() );
			System.err.print(", rot. vel. = "); System.err.println( vel.getY() );
		}
		int friendsNb = 0;
		for(int iDetect = 0; (perceptions != null) 
				&& (iDetect < perceptions.length); iDetect++) {
			final OrientPoint percCfg = perceptions[iDetect].config(); 
			final Vector2D percPos = percCfg.position();
			final Color    percCol = perceptions[iDetect].color();
			// --- avoiding collisions ----------------------------------------
			/* S'il y a collision, ajoutez :
			 * collision = 5; 
			 * System.err.println("---=# !!! COLLISION !!! #=---");
			 * sinon :
			 * Vector2D add;  // Calculez le vecteur d'évitement
			 * vectors[eVectors.avoidance.ordinal()].add(add); 
			 */
			// --- cohesion and alignment (not for leaders) -----------------
			if ( (! isLeader) && 
				 (perceptions[iDetect].color().hashCode() == robotColor.hashCode() ) ) {
				friendsNb++;
				/* À faire :
				vectors[eVectors.cohesion.ordinal()].add(); 
				vectors[eVectors.alignment.ordinal()].add(); 
				*/
			}
		}
		
		if (debug) { 
			System.err.print("Number of friends = "); System.err.println(friendsNb);
		}
		// nothing added for the cohesion
		// (position = origin in the robot's frame)
		// normalises cohesion and alignment
		if (friendsNb > 0) {
		   	// alignement should take into account the robot's 
		   	// orientation (0 in its own frame)
		   	vectors[eVectors.alignment.ordinal()].add
					( Vector2D.fromPolar(1, 0) );
		    	friendsNb++;  // take into account the robot
			final double friendsFact = 1. / friendsNb;
			vectors[eVectors.alignment.ordinal()].mult(friendsFact);
			vectors[eVectors.cohesion.ordinal()].mult(friendsFact);
		}
		// combines the vectors
		Vector2D combvectors = new Vector2D();
		for(int iVectors = 0; iVectors < eVectors.vectorsNb.ordinal(); iVectors++) 
			combvectors.add( vectors[iVectors].mult(vectorsCoef[iVectors]) ); 
		// set the goal velocity (random if no perceptions)
		Vector2D goalVel = (combvectors.length() < diameter / 10) 
				? new Vector2D( linVelMax * (Math.random() * 0.5 + 0.25),
						        rotVelMax * (Math.random() * 1.5 - 0.75) ) 
				: combvectors.mult(linVelMax / diameter / 2);
		// bound the goal vector (& inverse orientation variation)
		// !!! linear velocity should remain positive (no backup manoeuvre) !!!
		// (how can a robot follow a leader going backward?)
		goalVel.set( Math.max(Math.min(goalVel.getX(), linVelMax), 0),
				 	 Math.max(Math.min(goalVel.getY(), rotVelMax), -rotVelMax) ); 
		if (debug) { 
			System.err.print("--- Velocity goal = ("); System.err.print( goalVel.getX() );
			System.err.print(", "); System.err.print( goalVel.getY() );
			System.err.println(")");
		} 
		// compute the chosen accelerations 
		Vector2D acc = goalVel.minus(vel).mult(1 / timeHoriz); 
		/* consider vector as a force does not work at all!!!
		if (combvectors.length() < accMax / 10) { // nearly no vector
			final double smallLinAcc = accMax / 3, smallRotVel = rotVelMax / 3, 
					rndLinAcc = (Math.random() * 2 - 1) * smallLinAcc, 
					rndRotVel = (Math.random() * 2 - 1) * smallRotVel;
			combvectors.add( new Vector2D(rndLinAcc, rndRotVel) );  // add random
		}
		final double goalRot = combvectors.getY(), dRot = goalRot - vel.getY();
		Vector2D acc = new Vector2D(combvectors.getX(), dRot / timeHoriz); */
		final int linAccFact = acc.getX() < 0 ? -1 : 1, 
				rotAccFact = acc.getY() < 0 ? -1 : 1;
		// bounds the acceleration (not enough for the rotation?)
		acc.set( linAccFact * Math.max(Math.min(linAccFact * acc.getX(), linAccFact == linFact
									   		    ? linAccMax : accMax), accMin),
				 rotAccFact * Math.max(Math.min(rotAccFact * acc.getY(), rotAccFact == rotFact
						   					    ? rotAccMax : rotMaxAcc), rotMinAcc) );
		/*
		 new Vector2D( linFact * ( accMin + Math.random() * (linAccMax - accMin) ), 
					   rotFact * ( rotMinAcc + Math.random() * (rotAccMax - rotMinAcc) ) );
				*/ 
		if (debug) { 
			System.err.print("          acc = ("); System.err.print( acc.getX() );
			System.err.print(", "); System.err.print( acc.getY() );
			System.err.println(") ---");
		} 
		// send it to the environment
		robotWorld.setInfluence(id, acc);
    }
    
    public void move(Vector2D acc) { 
		final boolean debug = false;  // activates the console output
    	// verifies whether acceleration and velocity bounds are respected
		final double linFact = vel.getX() < 0 ? -1 : 1, linAcc = acc.getX(), 
				rotFact = vel.getY() < 0 ? -1 : 1, rotAcc = acc.getY();
		if (vel.getX() * linAcc > 0) {
			if (linFact * linAcc > accMax) {
				System.err.print("---=# !! Linear acceleration is too big (");
				System.err.print(linAcc); System.err.print(" / "); 
				System.err.print(accMax); System.err.println(") !!! #=---");
				acc.set(linFact * accMax, rotAcc);
		} } else {
			if (linFact * linAcc > -accMin)  {
				System.err.print("---=# !! Linear deceleration is too big (");
				System.err.print(linAcc); System.err.print(" / "); 
				System.err.print(-accMin); System.err.println(") !!! #=---");
				acc.set(-linFact * accMin, rotAcc);
		} }
		if (vel.getY() * rotAcc > 0) {
			if (rotFact * rotAcc > accMax / radius) {
				System.err.print("---=# !! Rotation acceleration is too big (");
				System.err.print(rotAcc); System.err.print(" / "); 
				System.err.print(accMax / radius); System.err.println(") !!! #=---");
				acc.set(linAcc, rotFact * accMax / radius);
		} } else {
			if (rotFact * rotAcc > -accMin / radius)  {
				System.err.print("---=# !! Rotation deceleration is too big (");
				System.err.print(rotAcc); System.err.print(" / "); 
				System.err.print(-accMin / radius); System.err.println(") !!! #=---");
				acc.set(linAcc, -rotFact * accMin / radius);
		} }
		if (linFact * (vel.getX() + linAcc * Environment.timeStep) > linVelMax) {
			System.err.println("---=# !! Linear velocity gets too big !!! #=---");
			acc.set( ( linFact * linVelMax - vel.getX() ) / Environment.timeStep, rotAcc );
		}
		if (rotFact * (vel.getY() + rotAcc * Environment.timeStep) > rotVelMax) {
			// needs correction!!!
			//System.err.println("---=# !! Rotation velocity gets too big !!! #=---");
			acc.set(linAcc, ( rotFact * rotVelMax - vel.getY() ) / Environment.timeStep);
		}
		if (debug) {
	    	System.err.print("("); System.err.print( config().position().getX() );
	    	System.err.print(", "); System.err.print( config().position().getY() );
	    	System.err.print(", "); System.err.print( config().orientDeg() );
	    	System.err.print(", "); System.err.print( vel.getX() );
	    	System.err.print(", "); System.err.print( OrientPoint.rad2deg( vel.getY() ) );
	    	System.err.print(") + ("); System.err.print( acc.getX() );
	    	System.err.print(", "); System.err.print( OrientPoint.rad2deg( acc.getY() ) ); 
		}
    	// small steps for an integration approximation
    	final int nbSteps = 100;
    	final double dt = Environment.timeStep / nbSteps;
    	final Vector2D dVel = acc.mult(dt);
    	for(int iT = 0; iT < nbSteps; iT++)
    	{
    		final Vector2D newVel = vel.plus(dVel), averageVel = vel.plus(newVel).mult(.5);
    		final double orient = config().orientation(), dist = averageVel.getX() * dt,
    				dOrient = averageVel.getY() * dt, averageOrient = orient + dOrient / 2;
    		// moving along a small segment of average orientation
    		config().add(Vector2D.fromPolar(dist, averageOrient), dOrient);
    		vel = newVel;
    	} 
    	if (debug) {
		    System.err.print(") +-> ("); System.err.print( config().position().getX() );
	    	System.err.print(", "); System.err.print( config().position().getY() );
	    	System.err.print(", "); System.err.print( config().orientDeg() );
	    	System.err.print(", "); System.err.print( vel.getX() );
	    	System.err.print(", "); System.err.print( OrientPoint.rad2deg( vel.getY() ) ); 
	    	System.err.println(")"); 
	    }
    }
}

