/** A vector in R^2, used to represent two dimension points or vectors.
 *
 * This class is needed as java.awt.geom.Point2D.Double is far 
 * too simple (not enough methods).
 *
 * @author  A. Scheuer, UL, FST, DI and LORIA.
 */
public class Vector2D {
    private double x, y;

    // === Constructors and related methods ==============================
    
    /** Default constructor, defining a (0, 0) vector. */
    public Vector2D() { set(0, 0); }
    
    /** Standard constructor, defining a (x, y) vector.
     * @param x  the abscissa of the vector,
     * @param y  the ordinate of the vector. 
     * @see   #getX()  @see #getY()  @see #set(double, double) */
    public Vector2D(final double x, final double y) { set(x, y); }
    
    /** The copy constructor.
     * @param v  the vector copied into the current one. 
     * @see   #getX()  @see #getY()  @see #set(double, double) */
    public Vector2D(final Vector2D v) { set( v.getX(), v.getY() ); }
    
    /** A static method to built a vector from its polar coordinates.
     * @param length  the length of the vector, 
     * @param angle   the angle of the vector.
     * @return  the vector of given length and angle. 
     * @see #length()  @see #angle() */
    public static Vector2D fromPolar(final double length, final double angle)
    { return ( new Vector2D( Math.cos(angle), Math.sin(angle) ) ).mult(length); }

    // === Description methods ===========================================
    
    /** The abscissa of the vector. 
     * @return  the abscissa of the vector. */
    public double getX() { return x; };
    
    /** The ordinate of the vector.  
     * @return  the ordinate of the vector. */
    public double getY() { return y; };

    /** The square length of the vector.  
     * @return  the square length of the vector.
     * @see     #scalProd(Vector2D)  @see #length() */
    public double sqrLength() { return scalProd(this); }

    /** The length of the vector. 
     * @return  the length of the vector.
     * @see  #sqrLength() */
    public double length() { return Math.sqrt( sqrLength() ); }

    /** The angle of the vector. 
     * @return  the angle of the vector.
     * @see  Math#atan2(double, double) */
    public double angle() { return Math.atan2( getY() , getX() ); }

    // === Modification methods ==========================================
    
    /** Main modification method, setting the vector to (x, y).
     * @param x  the future abscissa of the vector,
     * @param y  the future ordinate of the vector. */
    public void set(final double x, final double y) 
    { this.x = x; this.y = y; }

    /** Addition method, modifying the current vector.
     * @param v  the added vector.  @see #set(double, double) */
    public void add(final Vector2D v)
    { set( getX() + v.getX(), getY() + v.getY() ); }

    /** Multiplication method, modifying the current vector.
     * @param factor  the multiplication factor.  @see #set(double, double) */
    public void times(final double factor)
    { set( factor * getX(), factor * getY() ); }

    /** Subtraction method, modifying the current vector. 
     * @param v  the subtracted vector. 
     * @see #add(Vector2D)  @see #opposite() */
    public void subs(final Vector2D v) { add( v.opposite() ); }

    /** Rotation method, modifying the current vector. 
     * @param angle  the rotation angle.  @see  #angle(). */
    public void turn(final double angle) { 
    	final Vector2D v = fromPolar(1, -angle);
    	set( scalProd(v), -vectProd(v) );
    }
    
    // === Operation methods (not modifying the current vector) ==========

    /** Addition operation (the current vector remains constant).
     * @param v  the added vector. 
     * @return  the addition of the current vector and the given one. 
     * @see #add(Vector2D) */
    public Vector2D plus(final Vector2D v)
    { Vector2D res = new Vector2D(this); res.add(v); return res; }

    /** Multiplication operation (the current vector remains constant).
     * @param factor  the multiplication factor.  
     * @return  the multiplication of the current vector by the given factor. 
     * @see #times(double) */
    public Vector2D mult(final double factor) { 
		Vector2D res = new Vector2D(this);
		res.times(factor); return res;
    }

    /** Opposition operation (the current vector remains constant).  
     * @return  the opposite of the current vector.  @see #mult(double) */
    public Vector2D opposite() { return mult(-1.); }

    /** Subtraction operation (the current vector remains constant). 
     * @param v  the subtracted vector. 
     * @return   the subtraction of the given vector from the current one. 
     * @see #subs(Vector2D)  @see #plus(Vector2D)  @see #opposite() */
    public Vector2D minus(final Vector2D v)
    { return plus( v.opposite() ); }

    /** Rotation operation (the current vector remains constant). 
     * @param angle  the rotation angle. 
     * @return  the rotation of the current vector by the given angle. 
     * @see  #turn(double)  @see #angle() */
    public Vector2D rotate(final double angle) { 
		Vector2D res = new Vector2D(this);
		res.turn(angle); return res;
    }

    /** Scalar product between the current vector and a given one. 
     * @param v  an other vector. 
     * @return  the scalar product between the two vectors. */
    public double scalProd(final Vector2D v)
    { return getX() * v.getX() + getY() * v.getY(); }

    /** Vectorial product between the current vector and a given one. 
     * @param v  an other vector. 
     * @return  the vectorial product between the current vector and the given one. 
     */
    public double vectProd(final Vector2D v)
    { return getX() * v.getY() - getY() * v.getX(); }

    /** Descriptive method, printing the vector into a stream.
     * @param   out  the stream into which the vector will be printed. */
    public void print(java.io.PrintStream out) {
    	out.print("("); out.print( getX() ); out.print(", "); 
    	out.print( getY() ); out.print(")");
    }
    
} // end of class Vector2D
